"""
Support for sequentially writing uncompressed .npz files.
"""

__all__ = ["NpzWriter"]

from pathlib import Path
from typing import Literal, Union
from zipfile import ZipFile

import numpy as np
from numpy.typing import ArrayLike


class NpzWriter(object):
    """
    Helper class to sequentially add arrays to a given uncompressed .npz file.

    Parameters
    ----------
    fn : str or Path
        The zipped archive to replace, create, or append to.
    mode : str, optional
        The mode can be either write 'w', exclusive create 'x', or append 'a'.
        It defaults to 'x', so it will fail for existing files. Use 'w' to
        replace existing files, or 'a' to append to existing files.

    Examples
    --------
    Store data sequentially on disk, possibly overwriting an existing .npz file:

    >>> with mmnpz.NpzWriter("/tmp/123.npz", mode="w") as f:
    >>>     f.write("a", [[1, 2], [3, 4]])
    >>>     f.write("b", [5, 6, 7, 8])

    Append more data later:

    >>> with mmnpz.NpzWriter("/tmp/123.npz", mode="a") as f:
    >>>     f.write("c", [[9.5], [10.5]])

    Attributes
    ----------
    zip : ZipFile instance
        The ZipFile object for accessing the .npz file.
    """

    def __init__(self, fn: Union[str, Path], mode: Literal["w", "x", "a"] = "x") -> None:
        if mode not in ("w", "x", "a"):
            raise ValueError("NpzWriter requires mode 'w', 'x' or 'a', got %r" % mode)
        self.zip = ZipFile(fn, mode=mode)

    def write(self, name: str, array: ArrayLike) -> None:
        """
        Writes the given array under the given name.

        Parameters
        ----------
        name : str
            File name in the archive to save as, without its ``.npy`` extension.
            Supports directories separated by slashes.
        array : array-like
            Array to save to the file.
        """
        array = np.asanyarray(array)
        with self.zip.open(name + ".npy", "w", force_zip64=True) as f:
            np.lib.format.write_array(f, array, allow_pickle=False)

    def close(self) -> None:
        """
        Close the file.
        """
        if hasattr(self, "zip"):
            self.zip.close()
            del self.zip

    def __enter__(self) -> "NpzWriter":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    def __del__(self):
        self.close()



"""
Support for reading memory-mapped uncompressed .npz files.
"""

__all__ = ["NpzReader", "load"]

import struct
from collections.abc import Mapping
from pathlib import Path
from typing import Dict, Iterator, List, Literal, Tuple, Union
from zipfile import ZipFile

import numpy as np
from typing_extensions import Buffer


class NpzReader(Mapping):
    """
    A dictionary-like object with lazy-loading of numpy arrays in the given
    uncompressed .npz file. Upon construction, creates a memory map of the
    full .npz file, returning views for the arrays within on request. Serves
    as a drop-in replacement for :func:`numpy.load`.

    Parameters
    ----------
    fn : str or Path
        The zipped archive to open.
    mmap_mode : str, optional
        Provided for compatibility with numpy. Only supports 'r', indicating
        that the file is loaded as a readonly memory map.
    cache : bool, optional
        Whether to cache array objects in case they are requested again.
    preload : bool, optional
        Whether to precreate all array objects upon opening. Enforces caching.

    Raises
    ------
    OSError
        If the input file does not exist or cannot be read.
    zipfile.BadZipFile
        If the input file cannot be interpreted as a zip file.

    Examples
    --------
    Store data to disk, and load it again as a memory map:

    >>> np.savez("/tmp/123.npz", a=np.array([[1, 2], [3, 4]]))
    >>> data = mmnpz.NpzReader("/tmp/123.npz")
    >>> data["a"]
    memmap([[1, 2],
            [3, 4]])

    Attributes
    ----------
    files : list of str
        List of all uncompressed files in the archive with a ``.npy`` extension
        (listed without the extension). These are supported as dictionary keys.
    mmap : numpy.memmap
        The memory map of the full .npz file.
    arrays : dict
        Preloaded or cached arrays.
    """

    def __init__(
        self,
        fn: Union[str, Path],
        mmap_mode: Literal["r"] = "r",
        *,
        cache: bool = True,
        preload: bool = False,
    ) -> None:
        if mmap_mode != "r":
            raise ValueError("NpzReader requires mmap_mode='r', got %r" % mmap_mode)
        with ZipFile(fn, mode="r") as f:
            self._offsets: Dict[str, Tuple[int, int]] = {
                zinfo.filename[:-4]: (zinfo.header_offset, zinfo.file_size)
                for zinfo in f.infolist()
                if zinfo.filename.endswith(".npy") and zinfo.compress_type == 0
            }
        self.files: List[str] = list(self._offsets.keys())
        self.mmap = np.memmap(fn, mode="r")
        self.cache = cache or preload
        self.preload = preload
        self.arrays: Dict[str, np.memmap]
        if self.preload:
            self.arrays = {name: self.load(name) for name in self.files}
        else:
            self.arrays = {}

    def load(self, name: str) -> np.memmap:
        """
        Creates a view for the specified array, disregarding the cache.

        Parameters
        ----------
        name : str
            File name in the archive to load, without its ``.npy`` extension.

        Returns
        -------
        numpy.memmap
            A view into the global memory map with correct shape and dtype.

        See also
        --------
        __getitem__ : Returns a view for the specified array using the cache.
        """
        header_offset, file_size = self._offsets[name]
        # parse lengths of local header file name and extra fields
        # (ZipInfo is based on the global directory, not local header)
        fn_len, extra_len = struct.unpack("<2H", self.mmap[header_offset + 26 : header_offset + 30])
        # compute offset of start and end of data
        npy_start = header_offset + 30 + fn_len + extra_len
        npy_end = npy_start + file_size
        # read NPY header
        fp = MemoryviewIO(self.mmap)
        fp.seek(npy_start)
        version = np.lib.format.read_magic(fp)
        np.lib.format._check_version(version)
        shape, fortran, dtype = np.lib.format._read_array_header(fp, version)
        # produce slice of memmap
        data_start = fp.tell()
        return (
            self.mmap[data_start:npy_end]
            .view(dtype=dtype)
            .reshape(shape, order="F" if fortran else "C")
        )

    def close(self) -> None:
        """
        Releases the memory map and any cached views.
        """
        if hasattr(self, "mmap"):
            del self.mmap
        self.arrays = {}

    def __enter__(self) -> "NpzReader":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    def __iter__(self) -> Iterator[str]:
        return iter(self.files)

    def __len__(self) -> int:
        return len(self.files)

    def __getitem__(self, key: str) -> np.memmap:
        if self.cache:
            try:
                return self.arrays[key]
            except KeyError:
                pass
        array = self.load(key)
        if self.cache:
            self.arrays[key] = array
        return array

    def __contains__(self, key: object) -> bool:
        # Mapping.__contains__ calls __getitem__, which could be expensive
        return key in self._offsets


load = NpzReader  # Alias for serving as a :func:`numpy.load` replacement


class MemoryviewIO(object):
    """
    Wraps an object supporting the buffer protocol to be a readonly file-like.
    """

    def __init__(self, buffer: Buffer):
        self._buffer = memoryview(buffer).cast("B")
        self._pos = 0
        self.seekable = lambda: True
        self.readable = lambda: True
        self.writable = lambda: False

    def seek(self, offset: int, whence: int = 0):
        if whence == 0:
            self._pos = offset
        elif whence == 1:
            self._pos += offset
        elif whence == 2:
            self._pos = self._buffer.nbytes + offset

    def read(self, size: int = -1) -> bytes:
        data = self._buffer[self._pos : self._pos + size if size >= 0 else None].tobytes()
        self._pos += len(data)
        return data

    def tell(self) -> int:
        return self._pos



import zipfile

import numpy as np
import pytest

import mmnpz


@pytest.fixture(scope="session")
def good_npz(tmp_path_factory):
    fn = tmp_path_factory.mktemp("data") / "good.npz"
    np.savez(fn, a=np.array([[1, 2], [3, 4]]), b=np.linspace(0, 1, 6))
    return fn


@pytest.fixture(scope="session")
def bad_npz(tmp_path_factory):
    fn = tmp_path_factory.mktemp("data") / "bad.npz"
    with open(fn, "wb") as f:
        f.write(b"bad")
    return fn


@pytest.fixture(scope="session")
def absent_npz(tmp_path_factory):
    return tmp_path_factory.mktemp("data") / "absent.npz"


@pytest.mark.parametrize("preload", [False, True])
@pytest.mark.parametrize("cache", [False, True])
def test_read_success(good_npz, cache, preload):
    numpy = np.load(good_npz)
    ours = mmnpz.NpzReader(good_npz, cache=cache, preload=preload)
    assert len(numpy) == len(ours)
    assert numpy.files == list(ours)
    assert numpy.files == ours.files
    for key in numpy:
        assert key in ours
        assert isinstance(ours[key], np.memmap)
        assert np.array_equal(numpy[key], ours[key])
    assert "them" not in ours


def test_read_fail(bad_npz, absent_npz):
    with pytest.raises(zipfile.BadZipFile):
        mmnpz.NpzReader(bad_npz)
    with pytest.raises(IOError):
        mmnpz.NpzReader(absent_npz)


def test_cache(good_npz):
    data = mmnpz.NpzReader(good_npz)
    assert len(data.arrays) == 0
    a = data["a"]
    assert data.arrays == {"a": a}
    b = data["b"]
    assert data.arrays == {"a": a, "b": b}
    # check that repeated access gives the same view
    assert data["a"] is a
    assert data["b"] is b


def test_preload(good_npz):
    numpy = np.load(good_npz)
    ours = mmnpz.NpzReader(good_npz, preload=True)
    ours_preloaded = ours.arrays
    for key in numpy:
        assert key in ours_preloaded
        assert np.array_equal(numpy[key], ours_preloaded[key])


def test_nocache(good_npz):
    data = mmnpz.NpzReader(good_npz, cache=False)
    # check that every access creates a new view
    assert data["a"] is not data["a"]


def test_mmap(good_npz):
    data = mmnpz.NpzReader(good_npz)
    assert isinstance(data.mmap, np.memmap)
    assert data["a"].base is data.mmap


def test_close(good_npz):
    data = mmnpz.NpzReader(good_npz)
    _ = data["a"]
    assert hasattr(data, "mmap")
    assert len(data.arrays) > 0
    data.close()
    assert not hasattr(data, "mmap")
    assert len(data.arrays) == 0


def test_context_manager(good_npz):
    with mmnpz.NpzReader(good_npz) as data:
        _ = data["a"]
    assert not hasattr(data, "mmap")
    assert len(data.arrays) == 0


def test_memoryviewio():
    m = mmnpz.reading.MemoryviewIO(b"abcd")
    assert m.tell() == 0
    assert m.read(1) == b"a"
    assert m.tell() == 1
    m.seek(0)
    assert m.tell() == 0
    assert m.read(1) == b"a"
    assert m.tell() == 1
    m.seek(2)
    assert m.tell() == 2
    assert m.read(1) == b"c"
    assert m.tell() == 3
    m.seek(-2, 1)
    assert m.tell() == 1
    assert m.read(2) == b"bc"
    assert m.tell() == 3
    m.seek(0, 2)
    assert m.tell() == 4
    assert m.read() == b""


def test_unsupported_mmap_mode(good_npz):
    with pytest.raises(ValueError):
        mmnpz.NpzReader(good_npz, mmap_mode="w")  # type: ignore



import numpy as np
import pytest

import mmnpz


def make_arrays(k: int = 5):
    for i in range(k):
        yield str(i), np.full(i, i)


def save_arrays(fn, *, k: int = 5, **kwargs):
    with mmnpz.NpzWriter(fn, **kwargs) as f:
        for key, array in make_arrays(k):
            f.write(key, array)


@pytest.fixture
def npz_file(tmp_path):
    fn = tmp_path / "test.npz"
    save_arrays(fn)
    return fn


def test_savez_success(npz_file):
    with np.load(npz_file) as data:
        for key, array in make_arrays():
            assert key in data
            assert np.array_equal(data[key], array)


def test_exclusive(npz_file, tmp_path):
    with pytest.raises(IOError):
        with mmnpz.writing.NpzWriter(npz_file):
            pass
    with mmnpz.writing.NpzWriter(tmp_path / "test2.npz"):
        pass


def test_overwrite(npz_file):
    save_arrays(npz_file, k=3, mode="w")
    with np.load(npz_file) as data:
        assert len(data) == 3
        for key, array in make_arrays(3):
            assert key in data
            assert np.array_equal(data[key], array)


def test_append(npz_file):
    with mmnpz.writing.NpzWriter(npz_file, mode="a") as f:
        f.write("5", np.full(5, 5))
    with np.load(npz_file) as data:
        for key, array in make_arrays(6):
            assert key in data
            assert np.array_equal(data[key], array)


def test_unsupported_mode(tmp_path):
    with pytest.raises(ValueError):
        mmnpz.writing.NpzWriter(tmp_path / "test.npz", mode="r")  # type: ignore
